// Geometric Tools, Inc.
// http://www.geometrictools.com
// Copyright (c) 1998-2006.  All Rights Reserved
//
// The Wild Magic Library (WM3) source code is supplied under the terms of
// the license agreement
//     http://www.geometrictools.com/License/WildMagic3License.pdf
// and may not be copied or disclosed except in accordance with the terms
// of that agreement.

#ifndef WM3QUERY2TRATIONAL_H
#define WM3QUERY2TRATIONAL_H

#include "Wm3FoundationLIB.h"
#include "Wm3Query2.h"
#include "Wm3RVector2.h"

namespace Wm3
{

template <class Real>
class Query2TRational : public Query2<Real>
{
public:
    // The components of the input vertices are represented exactly as
    // rational values.
    Query2TRational (int iVQuantity, const Vector2<Real>* akVertex);
    virtual ~Query2TRational ();

    // run-time type information
    virtual Query::Type GetType () const;

    // Queries about the relation of a point to various geometric objects.

    // returns
    //   +1, on right of line
    //   -1, on left of line
    //    0, on the line
    virtual int ToLine (int i, int iV0, int iV1) const;
    virtual int ToLine (const Vector2<Real>& rkP, int iV0, int iV1) const;

    // returns
    //   +1, outside triangle
    //   -1, inside triangle
    //    0, on triangle
    virtual int ToTriangle (int i, int iV0, int iV1, int iV2) const;
    virtual int ToTriangle (const Vector2<Real>& rkP, int iV0, int iV1,
        int iV2) const;

    // returns
    //   +1, outside circumcircle of triangle
    //   -1, inside circumcircle of triangle
    //    0, on circumcircle of triangle
    virtual int ToCircumcircle (int i, int iV0, int iV1, int iV2) const;
    virtual int ToCircumcircle (const Vector2<Real>& rkP, int iV0, int iV1,
        int iV2) const;

private:
    using Query2<Real>::m_iVQuantity;
    using Query2<Real>::m_akVertex;

    // Caching for rational representations of the input.  The conversion of
    // floating-point numbers to TRational form is slow, so it is better to
    // keep track of which values have been converted.
    typedef TRational<4*sizeof(Real)> Rational;
    typedef RVector2<4*sizeof(Real)> RVector;
    mutable RVector* m_akRVertex;
    mutable bool* m_abEvaluated;

    void Convert (int iQuantity, int* aiIndex) const;

    int ToLine (const RVector& rkRP, int iV0, int iV1) const;
    int ToTriangle (const RVector& rkRP, int iV0, int iV1, int iV2) const;
    int ToCircumcircle (const RVector& rkRP, int iV0, int iV1, int iV2) const;

    static Rational Dot (Rational& rkX0, Rational& rkY0, Rational& rkX1,
        Rational& rkY1);

    static Rational Det2 (Rational& rkX0, Rational& rkY0, Rational& rkX1,
        Rational& rkY1);

    static Rational Det3 (Rational& rkX0, Rational& rkY0, Rational& rkZ0,
        Rational& rkX1, Rational& rkY1, Rational& rkZ1, Rational& rkX2,
        Rational& rkY2, Rational& rkZ2);
};

#include "Wm3Query2TRational.inl"

typedef Query2TRational<float> Query2TRationalf;
typedef Query2TRational<double> Query2TRationald;

}

#endif

